#!/usr/bin/env python3
#!python

import sys
import argparse
import os

def main():
    known_args, other_args = get_cmdline_options()
    create_mangled_smt2_file(known_args)

def get_cmdline_options():
    parser = argparse.ArgumentParser(description="A script for reducing OMT(FP) into OMT(BV)")
    parser.add_argument("smt2_file", metavar="<formula>.smt2", type=str,
                        help="The SMT-LIBv2 formula", action=check_ext(['smt2']))
    parser.add_argument("--bv_goal", metavar="<bv_name>", type=str,
                        help="unique name for support bv variable", default="my_new_bv_support")
    parser.add_argument("--new_goal", metavar="<new_name>", type=str,
                        help="unique name for new optimization goal", default="my_new_bv_goal")
    return parser.parse_known_args()


def create_mangled_smt2_file(opts):
    direction, goal = get_goal(opts.smt2_file)
    exp_width, sig_width = get_goal_decl(opts.smt2_file, goal)
    with open(opts.smt2_file, "r") as fd_in:
        for line in fd_in:
            if not line.startswith("(minimize") and not line.startswith("(maximize"):
                if "(get-objectives" in line:
                    print("(get-value ({}))".format(goal))
                    print(line)
                else:
                    print(line, end='')
            else:
                if direction == "maximize":
                    s = get_maximize_goal_string(opts, goal, exp_width, sig_width)
                elif direction == "minimize":
                    s = get_minimize_goal_string(opts, goal, exp_width, sig_width)
                print(s, end='')
    return


def check_ext(choices):
    """Checks that the argument extension matches the given list of choices."""
    class Act(argparse.Action): # pylint: disable=too-few-public-methods
        """Act."""
        def __call__(self, parser, namespace, fname, option_string=None):
            """Check that the argument extension matches the given list of choices"""
            ext = os.path.splitext(fname)[1][1:]
            if ext not in choices:
                option_string = '({})'.format(option_string) if option_string else ''
                parser.error("file doesn't end with one of {}{}".format(choices, option_string))
            else:
                setattr(namespace, self.dest, fname)
    return Act

def get_goal(smt2_file):
    # does not support :lower, :upper and :id
    # it assumes that there is only one goal
    with open(smt2_file, "r") as fd_in:
        for line in fd_in:
            if not line.startswith("(minimize") and not line.startswith("(maximize"):
                continue
            else:
                direction, goal = line[1:-2].split()
                return direction, goal

def get_goal_decl(smt2_file, goal):
    match = "(declare-fun {}".format(goal)

    with open(smt2_file, "r") as fd_in:
        for line in fd_in:
            if match not in line:
                continue
            if "FloatingPoint" not in line:
                if "Float32" in line:
                    return 8, 24
                else:
                    name = line.replace(')', ' ').split()[-1:]
                    match = "(define-sort {}".format(name[0])
                    fd_in.seek(0)
                    continue
                    break
            exp_width, sig_width = line.replace(')', ' ').split()[-2:]
            return int(exp_width), int(sig_width)

    with open(smt2_file, "r") as fd_in:
        for line in fd_in:
            if match not in line:
                continue
            assert "FloatingPoint" not in line
            exp_width, sig_width = line.replace(')', ' ').split()[-2:]
            return int(exp_width), int(sig_width)

def get_maximize_goal_string(opts, fp_goal, exp_width, sig_width):
    args = {
        'fp_goal' : fp_goal,
        'fp_e_size' : exp_width,
        'fp_s_size' : sig_width,
        'fp_size' : exp_width + sig_width,
        'bv_goal' : opts.bv_goal,
        'bv_size' : exp_width + sig_width,
        'new_goal' : opts.new_goal,
        'new_size' : exp_width + sig_width + 2,
        'fp_f' : exp_width + sig_width + 1,
        'fp_s' : exp_width + sig_width + 0,
        'fp_t' : exp_width + sig_width - 1,
        'bv_min_neg_mask' : '#b1' + ('1' * (exp_width + sig_width - 1)),
        'bv_min_pos_mask' : '#b0' + ('0' * (exp_width + sig_width - 1)),
        'bv_max_neg_mask' : '#b1' + ('0' * (exp_width + sig_width - 1)),
        'bv_max_pos_mask' : '#b0' + ('1' * (exp_width + sig_width - 1)),
        'bv_zero' : '#b0' + ('0' * (exp_width + sig_width - 1)),
    }
    return \
"""
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; support functions
(define-fun _bv_nxor ((x (_ BitVec {bv_size})) (y (_ BitVec {bv_size}))) (_ BitVec {bv_size})
    (bvor (bvand x y) (bvand (bvnot x) (bvnot y)))
)
; support variables
(declare-fun {bv_goal} () (_ BitVec {bv_size}))
(declare-fun {new_goal} () (_ BitVec {new_size}))
; constraints
(assert (= ((_ to_fp {fp_e_size} {fp_s_size}) {bv_goal}) {fp_goal}))
(assert (=> (= #b1 ((_ extract {fp_f} {fp_f}) {new_goal})) (not (fp.isNaN {fp_goal}))))
(assert (=> (= #b1 ((_ extract {fp_s} {fp_s}) {new_goal})) (fp.isPositive {fp_goal}))))
(assert (=  ((_ extract {fp_t} 0) {new_goal})
            (_bv_nxor {bv_goal}
                     (ite (= #b11 ((_ extract {fp_f} {fp_s}) {new_goal})) {bv_max_pos_mask}
                     (ite (= #b10 ((_ extract {fp_f} {fp_s}) {new_goal})) {bv_max_neg_mask}
                                                                          {bv_zero}
            )))
))
(maximize {new_goal})
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
""".format(**args)


def get_minimize_goal_string(opts, fp_goal, exp_width, sig_width):
    args = {
        'fp_goal' : fp_goal,
        'fp_e_size' : exp_width,
        'fp_s_size' : sig_width,
        'fp_size' : exp_width + sig_width,
        'bv_goal' : opts.bv_goal,
        'bv_size' : exp_width + sig_width,
        'new_goal' : opts.new_goal,
        'new_size' : exp_width + sig_width + 2,
        'fp_f' : exp_width + sig_width + 1,
        'fp_s' : exp_width + sig_width + 0,
        'fp_t' : exp_width + sig_width - 1,
        'bv_min_neg_mask' : '#b1' + ('1' * (exp_width + sig_width - 1)),
        'bv_min_pos_mask' : '#b0' + ('0' * (exp_width + sig_width - 1)),
        'bv_max_neg_mask' : '#b1' + ('0' * (exp_width + sig_width - 1)),
        'bv_max_pos_mask' : '#b0' + ('1' * (exp_width + sig_width - 1)),
        'bv_zero' : '#b0' + ('0' * (exp_width + sig_width - 1)),
    }
    return \
"""
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; support functions
(define-fun _bv_nxor ((x (_ BitVec {bv_size})) (y (_ BitVec {bv_size}))) (_ BitVec {bv_size})
    (bvor (bvand x y) (bvand (bvnot x) (bvnot y)))
)
; support variables
(declare-fun {bv_goal} () (_ BitVec {bv_size}))
(declare-fun {new_goal} () (_ BitVec {new_size}))
; constraints
(assert (= ((_ to_fp {fp_e_size} {fp_s_size}) {bv_goal}) {fp_goal}))

(assert (=> (= #b1 ((_ extract {fp_f} {fp_f}) {new_goal})) (not (fp.isNaN {fp_goal}))))
(assert (=> (= #b1 ((_ extract {fp_s} {fp_s}) {new_goal})) (fp.isNegative {fp_goal}))))
(assert (=  ((_ extract {fp_t} 0) {new_goal})
            (_bv_nxor {bv_goal}
                     (ite (= #b11 ((_ extract {fp_f} {fp_s}) {new_goal})) {bv_min_neg_mask}
                     (ite (= #b10 ((_ extract {fp_f} {fp_s}) {new_goal})) {bv_min_pos_mask}
                                                                          {bv_zero}
            )))
))
(maximize {new_goal})
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
""".format(**args)


###
###
###

if __name__ == "__main__":
    MIN_PYTHON = (3, 5)
    if sys.version_info < MIN_PYTHON:
        sys.exit("Python %d.%d or later is required.\n" % MIN_PYTHON)

    main()
