#!/bin/bash

## run_optimathsat:
#  - ${1} : fzn file
#  - ${2} : output file
#  - ${3} : 1 overwrite, 0 no overwrite
#  - ${4} : timeout
#  - ...  : options for optimathsat
#  returns ret

function run_optimathsat ()
{
    tool="optimathsat"
    msg=""
    ret=$((0))

    src_file=${1} ;
    dst_file=${2} ;
    overwrite=${3} ;
    timeout=${4} ;

    cmd_line="timeout --signal=SIGKILL \"${timeout}\" /usr/bin/time -v optimathsat ${@:5} < \"${src_file}\" 2>\"${dst_file}\" 1>&2"
    if [[ ${overwrite} -eq 1 || ! -f ${dst_file} ]]; then

        timeout --signal=SIGKILL "${timeout}" \
            /usr/bin/time -v \
                optimathsat "${@:5}" < "${src_file}" 2>"${dst_file}" 1>&2

        ret=$?

        if grep -q "msat error: sat, but with non-linear terms" "${dst_file}"; then
            ret=$((-101)) # force error - optimathsat
        elif grep -qi "error: division between two non-constant arguments unsupported" "${dst_file}"; then
            ret=$((-102)) # force error - optimathsat
        elif grep -qi "^line:" "${dst_file}"; then
            ret=$((-103)) # force error - catch all for optimathsat

        elif grep -qi "error: type error: .* type mismatch in line" "${dst_file}"; then
            ret=$((-111)) # force error - izplus
        elif grep -qi "error: .* not supported" "${dst_file}"; then
            ret=$((-112)) # force error - izplus
        elif grep -qi "error: registry: .* not found in line" "${dst_file}"; then
            ret=$((-113)) # force error - izplus

        elif grep -qi "<stdin>:.*: .* unsupported" "${dst_file}"; then
            ret=$((-121)) # force error -- hcsp
        elif grep -qi "<stdin>:.*: .* not supported" "${dst_file}"; then
            ret=$((-122)) # force error -- hcsp
        elif grep -qi "parsing failed" "${dst_file}"; then
            ret=$((-123)) # force error -- hcsp
        elif grep -qi "out of range" "${dst_file}"; then
            ret=$((-124)) # force error - hcsp

        elif grep -qi "fzn_interpretation_error" "${dst_file}"; then
            ret=$((-131)) # force error - picat sat

        elif grep -qi "error" "${dst_file}" | grep -qi "% error:" --invert-match ; then
            ret=$((-151)) # force error
        elif grep -qi "exception" "${dst_file}"; then
            ret=$((-152)) # force error
        fi

        if [[ ${ret} -eq 1 ]] || (( ret < 0 )); then
            OPTIMATHSAT_ERRORS=$((OPTIMATHSAT_ERRORS + 1))
            msg="\t${tool}(${RED}error${NORMAL}): ${dst_file}"
        elif [[ ${ret} -eq 124 ]]; then
            OPTIMATHSAT_TIMEOUTS=$((OPTIMATHSAT_TIMEOUTS + 1))
            msg="\t${tool}(${PINK}timeout${NORMAL}): ${dst_file}"
        elif [[ ${ret} -eq 137 ]]; then
            OPTIMATHSAT_KILLED=$((OPTIMATHSAT_KILLED + 1))
            msg="\t${tool}(${CYAN}killed${NORMAL}): ${dst_file}"
        elif [[ ${ret} -eq 139 ]]; then
            OPTIMATHSAT_SIGSEVS=$((OPTIMATHSAT_SIGSEVS + 1))
            msg="\t${tool}(${CYAN}sigsev${NORMAL}): ${dst_file}"
        elif [[ ${ret} -eq 0 ]]; then
            OPTIMATHSAT_OKS=$((OPTIMATHSAT_OKS + 1))
            msg="\t${tool}(${GREEN}ok${NORMAL}): ${dst_file}"
        else
            OPTIMATHSAT_UNKNOWNS=$((OPTIMATHSAT_UNKNOWNS + 1))
            msg="\t${tool}(${BLUE}unknown${NORMAL}): ${dst_file}"
        fi
    else
        OPTIMATHSAT_SKIPPED=$((OPTIMATHSAT_SKIPPED + 1))
        msg="\t${tool}(${YELLOW}skipped${NORMAL}) ${dst_file}"
    fi

    echo -e "\nEXEC RESULT (status: ${ret}):\n${msg}\n\t~$ ${cmd_line}" >> "${dst_file}"
    echo -e "${msg}"

    run_optimathsat="${dst_file}"
    return "${ret}"
}

#
# main
#

if [[ "${BASH_SOURCE[0]}" == "${0}" ]]; then
    run_optimathsat "${@}"
else
    :
fi
