#!/bin/bash

#
## FUNCTIONS
#
function apply_file () {	# applied over each .mzn file detected
	# src_file=${1} ;

	src_file=${1};

	if [[ ${src_file} =~ \.sat\.smt2 ]]; then
		return
	fi

	if [[ ${src_file} =~ \.unsat\.smt2 ]]; then
		return
	fi

	dst_file=$(echo ${src_file} \
			| sed 's#'${SRC_DIR}'#'${DST_DIR}'#' \
			| sed 's#\.smt2$#\.out#')

	test_formula "${src_file}" "${dst_file}"
	return
}
function apply_dir () {
    re=${1} ; func=${2} ;
    shift ; shift ;

	echo "Directory "`pwd`" ..."

	test $(( $(find . -maxdepth 1 -type d | wc -l) - 1)) -gt 0
	has_subdir=$? # 0 if has sub-dir, 1 otherwise

    test -n "$(find . -name '*.dzn')"
    mzn_only=$((1 - $?)) # 0 if there are only .mzn files, 1 otherwise

	for file in `ls -A ${re} 2>/dev/null`
	do
		eval '${func} `pwd`/${file} ${@} ${has_subdir} ${mzn_only}'
	done
}
function apply_rec () {
    pushd ${1} 1>/dev/null ;
    shift ;

    eval "${@}" ;

    for dir in `ls -d */ 2>/dev/null`
    do
        apply_rec `pwd`/${dir} ${@} ;
    done
    popd 1>/dev/null
}
function cp_dirtree () {
	dest_dir=`pwd | sed 's#'${SRC_DIR}'#'${DST_DIR}'#'`
	if [[ ! ( -d ${dest_dir} ) ]]; then
		mkdir ${dest_dir}
	fi
}

#
## PROBLEM HANDLERS
#


function test_formula()
{
	formula_count=$((formula_count + 1))

    fix_formula "${@}"
}

function fix_formula()
{
    local src_file="${1}"

    if grep -q "status unsat" "${src_file}"; then
        echo -e "\t(${PINK}warning${NORMAL}): status unsat in \"${src_file}\""
        return
    fi

    if [[ "${src_file}" =~ .*wintersteiger/abs/.* ]]; then
        fix_formula_wintersteiger_abs_min "${@}"
        fix_formula_wintersteiger_abs_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/add/.* ]]; then
        fix_formula_wintersteiger_add_min "${@}"
        fix_formula_wintersteiger_add_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/div/.* ]]; then
        fix_formula_wintersteiger_div_min "${@}"
        fix_formula_wintersteiger_div_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/eq/.* ]]; then
        fix_formula_wintersteiger_eq_min "${@}"
        fix_formula_wintersteiger_eq_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/gt/.* ]]; then
        fix_formula_wintersteiger_gt_min "${@}"
        fix_formula_wintersteiger_gt_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/lt/.* ]]; then
        fix_formula_wintersteiger_lt_min "${@}"
        fix_formula_wintersteiger_lt_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/max/.* ]]; then
        fix_formula_wintersteiger_max_min "${@}"
        fix_formula_wintersteiger_max_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/min/.* ]]; then
        fix_formula_wintersteiger_min_min "${@}"
        fix_formula_wintersteiger_min_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/mul/.* ]]; then
        fix_formula_wintersteiger_mul_min "${@}"
        fix_formula_wintersteiger_mul_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/rem/.* ]]; then
        fix_formula_wintersteiger_rem_min "${@}"
        fix_formula_wintersteiger_rem_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/sqrt/.* ]]; then
        fix_formula_wintersteiger_sqrt_min "${@}"
        fix_formula_wintersteiger_sqrt_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/sub/.* ]]; then
        fix_formula_wintersteiger_sub_min "${@}"
        fix_formula_wintersteiger_sub_max "${@}"
    elif [[ "${src_file}" =~ .*wintersteiger/toIntegral/.* ]]; then
        fix_formula_wintersteiger_toIntegral_min "${@}"
        fix_formula_wintersteiger_toIntegral_max "${@}"
    elif [[ "${src_file}" =~ .*griggio/.* ]]; then
        fix_formula_griggio_min "${@}"
        fix_formula_griggio_max "${@}"
    elif [[ "${src_file}" =~ .*schanda/.* ]]; then
        fix_formula_schanda_min "${@}"
        fix_formula_schanda_max "${@}"
    elif [[ "${src_file}" =~ .*ramalho/.* ]]; then
        fix_formula_ramalho_min "${@}"
        fix_formula_ramalho_max "${@}"
    elif [[ "${src_file}" =~ .*heizmann*/.* ]]; then
        fix_formula_heizmann_min "${@}"
        fix_formula_heizmann_max "${@}"
    else
        return
    fi

    return
}

##
## HEIZMANN
##
function fix_formula_heizmann_min()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    local idx=$((0))
    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.var_${idx}.min.smt2#")
        idx=$((idx + 1))

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(minimize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}
function fix_formula_heizmann_max()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    local idx=$((0))
    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.var_${idx}.max.smt2#")
        idx=$((idx + 1))

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(maximize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}

##
## RAMALHO
##
function fix_formula_ramalho_min()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    local idx=$((0))
    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.var_${idx}.min.smt2#")
        idx=$((idx + 1))

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)##" "${dst_file}"
            export cand="${cand}"
            perl -p -e 's/(check-sat)/minimize $ENV{cand})\n(check-sat)\n(get-objectives)\n(get-model/' "${dst_file}" > file.tmp
            mv file.tmp "${dst_file}"
#            sed -i 's?(check-sat)?(minimize '"${cand}"')\n(check-sat)\n(get-objectives)\n(get-model)?' "${dst_file}"
        fi
    done
}
function fix_formula_ramalho_max()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    local idx=$((0))
    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.var_${idx}.max.smt2#")
        idx=$((idx + 1))

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)##" "${dst_file}"
            export cand="${cand}"
            perl -p -e 's/(check-sat)/maximize $ENV{cand})\n(check-sat)\n(get-objectives)\n(get-model/' "${dst_file}" > file.tmp
            mv file.tmp "${dst_file}"
#            sed -i 's?(check-sat)?(maximize '"${cand}"')\n(check-sat)\n(get-objectives)\n(get-model)?' "${dst_file}"
        fi
    done
}

##
## SCHANDA
##
function fix_formula_schanda_min()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () Float" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.${cand}.min.smt2#")

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(minimize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}
function fix_formula_schanda_max()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () Float" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.${cand}.max.smt2#")

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(maximize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}

##
## GRIGGIO
##
function fix_formula_griggio_min()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.${cand/*::*/main}.min.smt2#")

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(minimize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}
function fix_formula_griggio_max()
{
    local src_file="${1}"
    local dst_base="${2}"

    # collect candidate objectives

    local candidates=()
    while read -r line ;
    do
        candidates+=($(echo ${line} | sed 's/.* \(.*\) () .*/\1/'))
    done < <(grep "(declare-fun .* () (_ FloatingPoint" ${src_file})

    if (( ${#candidates[@]} <= 0 )); then
        echo -e "\t(${YELLOW}warning${NORMAL}): no objective in \"${src_file}\""
    fi

    # generate formulas

    for cand in ${candidates[@]}
    do
        local dst_file=$(echo "${dst_base}" | sed "s#\.out#.${cand/*::*/main}.max.smt2#")

        if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
            cp "${src_file}" "${dst_file}"

            sed -i "s#(set-info :status .*)#(set-option :produce-models true)#" "${dst_file}"
            sed -i "s?(check-sat)?(maximize ${cand})\n(check-sat)\n(get-objectives)\n(get-model)?" "${dst_file}"
        fi
    done
}

##
## TOINTEGRAL
##
function fix_formula_wintersteiger_toIntegral_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_toIntegral_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## SUB
##
function fix_formula_wintersteiger_sub_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_sub_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        sed -i 's#= y#fp.geq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## SQRT
##
function fix_formula_wintersteiger_sqrt_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_sqrt_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## MUL
##
function fix_formula_wintersteiger_mul_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.geq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_mul_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## MIN
##
function fix_formula_wintersteiger_min_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.geq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_min_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## MAX
##
function fix_formula_wintersteiger_max_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.geq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_max_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## LT
##
function fix_formula_wintersteiger_lt_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        if grep -q "false))" "${dst_file}"; then
            sed -i 's#= x#fp.geq x#' "${dst_file}"
            sed -i 's#= y#fp.leq y#' "${dst_file}"
            sed -i 's#(check-sat)#(minimize x)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        else
            sed -i 's#= x#fp.leq x#' "${dst_file}"
            sed -i 's#= y#fp.geq y#' "${dst_file}"
            sed -i 's#(check-sat)#(minimize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        fi

        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_lt_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        if grep -q "false))" "${dst_file}"; then
            sed -i 's#= x#fp.leq x#' "${dst_file}"
            sed -i 's#= y#fp.geq y#' "${dst_file}"
            sed -i 's#(check-sat)#(maximize x)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        else
            sed -i 's#= x#fp.geq x#' "${dst_file}"
            sed -i 's#= y#fp.leq y#' "${dst_file}"
            sed -i 's#(check-sat)#(maximize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        fi

        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## GT
##
function fix_formula_wintersteiger_gt_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        if grep -q "true))" "${dst_file}"; then
            sed -i 's#= x#fp.geq x#' "${dst_file}"
            sed -i 's#= y#fp.leq y#' "${dst_file}"
            sed -i 's#(check-sat)#(minimize x)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        else
            sed -i 's#= x#fp.leq x#' "${dst_file}"
            sed -i 's#= y#fp.geq y#' "${dst_file}"
            sed -i 's#(check-sat)#(minimize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        fi

        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_gt_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        if grep -q "true))" "${dst_file}"; then
            sed -i 's#= x#fp.leq x#' "${dst_file}"
            sed -i 's#= y#fp.geq y#' "${dst_file}"
            sed -i 's#(check-sat)#(maximize x)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        else
            sed -i 's#= x#fp.geq x#' "${dst_file}"
            sed -i 's#= y#fp.leq y#' "${dst_file}"
            sed -i 's#(check-sat)#(maximize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        fi

        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## EQ
##
function fix_formula_wintersteiger_eq_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"

        sed -i 's#(assert (= y#;(assert (= y#' "${dst_file}"
        sed -i 's#false))#true))#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_eq_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"

        sed -i 's#(assert (= y#;(assert (= y#' "${dst_file}"
        sed -i 's#false))#true))#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize y)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## DIV
##
function fix_formula_wintersteiger_div_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}" # too much degree of freedom!

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_div_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        #sed -i 's#= y#fp.geq y#' "${dst_file}" # too much degree of freedom!

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## ADD
##
function fix_formula_wintersteiger_add_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.geq x#' "${dst_file}"
        sed -i 's#= y#fp.geq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_add_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#= x#fp.leq x#' "${dst_file}"
        sed -i 's#= y#fp.leq y#' "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

##
## ABS
##
function fix_formula_wintersteiger_abs_min()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.min.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r#' "${dst_file}"
        sed -i 's#(check-sat)#(minimize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}
function fix_formula_wintersteiger_abs_max()
{
    local src_file="${1}"
    local dst_base="${2}"
    local dst_file=$(echo "${dst_base}" | sed 's#\.out#.max.smt2#')

    if [[ ${OVERWRITE} -eq 1 || ! -f ${dst_file} ]]; then
        cp "${src_file}" "${dst_file}"

        sed -i 's#(assert (= r#;(assert (= r;#' "${dst_file}"
        sed -i 's#(check-sat)#(maximize r)\n(check-sat)\n(get-objectives)\n(get-model)#' "${dst_file}"
        sed -i 's#(set-info :status .*)#(set-option :produce-models true)#' "${dst_file}"
    fi
}

function reset_counters()
{
	formula_count=$((0))
}

function load_resources()
{
    true
}

## file containing result from time
#
function usr_time()
{
	usr_time="ERROR"

	grep -q "User time (seconds):" ${1} || return 1;

	usr_time=$(grep "User time (seconds):" ${1} | grep -oE "[[:digit:]]+\.[[:digit:]]+")
	return 0;
}

function sys_time()
{
	sys_time="ERROR"

	grep -q "System time (seconds):" ${1} || return 1;

	sys_time=$(grep "System time (seconds):" ${1} | grep -oE "[[:digit:]]+\.[[:digit:]]+")
	return 0;
}

function tot_time()
{
	tot_time="ERROR"

	usr_time ${1} || return 1;
	sys_time ${1} || return 1;

	tot_time=$(python -c "print ${usr_time} + ${sys_time}")
	return 0;
}

function result()
{
	result="ERROR"

	grep -q "^sat" ${1}     && result="SAT"
	grep -q "^unsat" ${1}   && result="UNSAT"
	grep -q "^unknown" ${1} && result="UNKNOWN"

	[[ "${result}" =~ ERROR ]] && return 1

	return 0;
}

function opt_value()
{
	opt_value="ERROR"

    opt_value=$(awk 'f&&NR==f+1; /objectives/ {f=NR}' "${1}" | cut -d\  -f3-5 | sed 's/)//')

    if [[ -z "${opt_value}" ]]; then
        opt_value="---"
    fi

	return 0;
}

#
## MAIN
#

## run_main
#	- ${1}: src_dir
#	- ${2}: dst_dir
#	- ${3}: overwrite
#	- ${4}: timeout
#
function run_main ()
{
	reset_counters

	export INIT_DIR=$(pwd)
	export SRC_REG='\*.smt2'
	export SRC_DIR="${1}"
	export DST_DIR="${2}"
	export OVERWRITE="${3}"
	export TIMEOUT="${4}"

	[ "${OVERWRITE}" ] || OVERWRITE="1" ;
	[ "${TIMEOUT}" ]   || TIMEOUT="0" ;

	[[ ! ( -d "${DST_DIR}" ) ]] && mkdir "${DST_DIR}"

	load_resources

	dirs -c
	apply_rec ${SRC_DIR} cp_dirtree
	apply_rec ${SRC_DIR} apply_dir ${SRC_REG} apply_file
}

if [[ "${BASH_SOURCE[0]}" == "${0}" ]]; then
	run_main "${@}"
else
	:
fi
