#!/usr/bin/env python

import os, sys, argparse
import numpy as np
import matplotlib
matplotlib.use('Agg') # fix matplotlib bug
import matplotlib.pyplot as plt
import re
from pylab import poly1d, xlim, ylim

###
### main
###


def main():
    opts = get_options()

    if len(opts.files) <= 0:
        print ("usage: stats_plot [-n title] [-d plots_dir] [-t timeout] stats_file ...")
        quit(1)

    configs = []
    results = {}
    for file in opts.files:
        collect_stats(file, configs, results, opts)

    good_data = check_integrity(configs, results)

    plot_table(configs, results)

    plots_dir = os.path.join(os.getcwd(), "plots")
    if not os.path.exists(plots_dir):
        os.makedirs(plots_dir)

    print ("Generating plots...")


    if not good_data:
        print ("missing data or inconsistent results, please fix the issue")
        quit(1)

    save_scatter_plots(plots_dir, configs, results, 600.0)
    save_cactus_plots(plots_dir, configs, results, 600.0)

###
### help functions
###

def get_options():
    """parses and returns input options"""
    parser = argparse.ArgumentParser(description="stats_plot")
    parser.add_argument("-n", type=str, help="plot name", default="default")
    parser.add_argument("-d", type=str, help="plots directory", default="plots")
    parser.add_argument("-t", type=int, help="timeout", default=600)
    parser.add_argument("files", type=str, nargs=argparse.REMAINDER)
    return parser.parse_args()

def collect_stats(file, configs, results, opts):
    """parses statistics summary file generated by run_experiment.sh, saving
    the interesting values in 'tools' and 'results'"""

    try:
        with open(file, 'r') as fd:
            file = os.path.realpath(file)
            config, ext = os.path.splitext(os.path.basename(file))
            if config not in configs:
                configs.append(config)

            for line in fd:
                idx, config_name, tool, res01, res02, res03, status, \
                time, timeout, smt2_file, out_file, opt_value \
                = line.split(" : ")[0:-1] #''.join(line.split()).split(":")[0:-1]

                if "bv" in opt_value:
                    # optimathsat and z3 print value differently
                    opt_value = opt_value.replace("bv", "").split()[1]

                key, ext = os.path.splitext(os.path.basename(smt2_file))

                if key not in results.keys():
                    results[key] = {}

                if int(res01) == 0 and (int(res02) != 0 or int(res03) != 0):
                    print("The following benchmark had an incorrect result: ", config_name, tool, smt2_file)
                    quit(1);

                if (time == "ERROR" and res01 == "137"):
                    time = opts.t
                    res01 = "124"
                    opt_value = ""

                if (float(time) <= 0.0):
                    time = float(time) + 0.01

                if int(res01) == 124 or int(res01) == 137:
                    time = opts.t
                elif (float(time) >= opts.t):
                    # in this experiment, the internal timeout of optimathsat is set to a higher
                    # value than the killing threshold of the outer script, since in many cases
                    # the internal timeout makes the process become a zombie. So the solver is
                    # allowed to execute for more time than it should, and those values that are
                    # above the threshold must be discarded manually.
                    time = opts.t
                    res01 = "124"
                    opt_value = ""

                results[key][config] = (float(time), int(res01), 0, \
                                       res01, opt_value)

    except Exception as e:
        print("error: file `" + file + "` does not exist or cannot be parsed, quitting.\n")
        print e
        quit(1)


def check_integrity(configs, results):
    """check that there are no obvious inconsistencies among solvers"""

    num_configs = len(configs)

    good = True

    for key in results.keys():
        num_results = len(results[key].keys())

        if (num_results < num_configs):
            print("warning: fewer results than expected for key `" + key + "`.")
            good = False
            pass

        v_results = []
        v_values = []

        for config in results[key].keys():
            data = results[key][config]

            if config == "mathsat_smt_only":
                # ignore SMT version
                continue

            if (data[1] != 0):
                # ignore entry, there is an error or timeout
                continue

            v_results.append(data[3])

            value = data[4]
            m = re.search('(fp #b1 #b00* #b00*)', value)
            if m is not None:
                value = value.replace(" #b1 ", " #b0 ")
            v_values.append(value)

        b_results = all(x == v_results[0] for x in v_results)
        b_values = all(x == v_values[0] for x in v_values)

        if not b_results or not b_values:
            print("error: inconsistent results for key `" + key + "`.")
            print("\tplease manually investigate the issue and change the tool")
            print("\texit status from 0 to 2 for the incorrect configuration / file")
            print results[key]
            good = False

    return good

def plot_table(configs, results):
    data = get_data_for_table(configs, results)

    print "%40s\t%5s\t%5s\t%5s\t%5s\t%5s\t%5s\t%10s\t%10s\t%10s\t%5s\t%5s\t%5s" % ("Config:", "Benchs:", "Solved:", "Timeouts:", "Unsupported:", "Unknown:", "Errors:", "Time (s.):", "Avg Term", "Avg All", "Unique", "Only best", "Shared best")

    for config in configs:
        values = (config, data[config]["benchs"], data[config]["solved"], data[config]["timeouts"], data[config]["unsupported"], data[config]["unknown"], data[config]["errors"], data[config]["cum_time"], data[config]["avg_term"], data[config]["avg_all"], data[config]["unique"], data[config]["only_best"], data[config]["shared_best"])
        print "%40s\t%5s\t%5s\t%5s\t%5s\t%5s\t%5s\t%10s\t%10s\t%10s\t%5s\t%5s\t%5s" % tuple(values)

def get_data_for_table(configs, results):

    ret = {}

    for config in configs:
        ret[config] = {
            "benchs"      : 0,
            "solved"      : 0,
            "timeouts"    : 0,
            "unsupported" : 0,
            "errors"      : 0,
            "unknown"     : 0,
            "cum_time"    : 0,
            "unique"      : 0,
            "only_best"   : 0,
            "shared_best" : 0
        }

    cc_only_best=0
    cc_shared_best=0

    for key in results.keys():
        # key is benchmark name
        solved = []
        for config in results[key].keys():
            if config in ["vb_all", "vb_lin", "vb_e_ofpbs"]:
                continue

            # config is configuration name
            ret[config]["benchs"] = ret[config]["benchs"] + 1
            
            es = results[key][config][1]

            has_error = False

            unsupported_list = [-101, 155, -102, 154, -103, 153, -111, 145, -112, 144, -113, 143, -121, 134, -122, 134, -123, 133, -124, 132, -131, 125, -151, 105]
            # generic error: -151, 105
            # exception: -152, 104
            # sigsev: -139
            # killed: 137

            if (es == 124 or es == 137):
                ret[config]["timeouts"] = ret[config]["timeouts"] + 1
            elif es in unsupported_list:
                ret[config]["unsupported"] = ret[config]["unsupported"] + 1
                #print("warning: unsupported result for `" + key + "`, " + config + ": " + str(es))
            elif es != 0:
                ret[config]["errors"] = ret[config]["errors"] + 1
                print("warning: error result for `" + key + "`, " + config + ":")
                print results[key][config]
            elif results[key][config][3] == "UNKNOWN":
                ret[config]["unknown"] = ret[config]["unknown"] + 1
                print("warning: unknown result for `" + key + "`, " + config + ":")
                print results[key][config]
            else:
                ret[config]["cum_time"] = ret[config]["cum_time"] + results[key][config][0]
                ret[config]["solved"] = ret[config]["solved"] + 1
                solved.append((config, results[key][config][0]))
        if len(solved) == 1:
            ret[config]["unique"] = ret[config]["unique"] + 1
        elif len(solved) > 1:
            solved.sort(key=lambda x: x[1])
            min_time = min(solved, key=lambda x: x[1])[1]
            filtered = [config for config, time in solved if time <= min_time]
            if len(filtered) == 1:
                ret[filtered[0]]["only_best"] = ret[filtered[0]]["only_best"] + 1
                cc_only_best += 1
            else:
                cc_shared_best += 1
                for config in filtered:
                    ret[config]["shared_best"] = ret[config]["shared_best"] + 1

    print("Only Best: " + str(cc_only_best))
    print("Shared Best: " + str(cc_shared_best))

    for config in configs:
        sum = ret[config]["solved"] + ret[config]["timeouts"] +    \
              ret[config]["unsupported"] + ret[config]["errors"] + \
              ret[config]["unknown"]
        if ret[config]["benchs"] != sum:
            print("error: incorrect value in columns")
            print ret[config]
            quit(1)
        ret[config]["avg_term"] = float(ret[config]["cum_time"]) / float(ret[config]["solved"])
        ret[config]["avg_all"] = float(ret[config]["cum_time"] +
                                 600.0 * (ret[config]["benchs"] - ret[config]["solved"])) \
                                 / float(ret[config]["benchs"])

    return ret

def get_xyz(x_tool, y_tool, results, timeout):
    x = []
    y = []
    z = []

    for bench in results.keys():
        x_data = results[bench][x_tool]
        y_data = results[bench][y_tool]

        error = False

        if x_data[1] != 0 and x_data[1] != 124:
            x.append(timeout)
            error = True
        else:
            x.append(x_data[0])


        if y_data[1] != 0 and y_data[1] != 124:
            y.append(timeout)
            error = True
        else:
            y.append(y_data[0])

        if error:
            z.append(0.5)
        else:
            if x_data[1] == 124 or y_data[1] == 124 or \
                x_data[0] >= timeout or y_data[0] >= timeout:
                z.append(1.0)
            else:
                z.append(0.0)

    return x, y, z

def save_scatter_plots(plots_dir, configs, results, timeout):
    for x_tool in configs:
        for y_tool in configs:
            if x_tool != y_tool:
                x, y, z = get_xyz(x_tool, y_tool, results, timeout)
                # args
                title   = "%s_%s"   % (x_tool, y_tool)
                x_label = "%s (s.)" % get_pretty_label(x_tool)
                y_label = "%s (s.)" % get_pretty_label(y_tool)
                z_label = ""
                file_name = "%s/%s.png" % (plots_dir, title)
                plot_scatter_plot(title, x_label, y_label, z_label, x, y, z, timeout, file_name)

                file_name = "%s/%s.pdf" % (plots_dir, title)
                plot_scatter_plot(title, x_label, y_label, z_label, x, y, z, timeout, file_name)

def plot_scatter_plot(title, x_label, y_label, z_label, x, y, z, timeout, file_name):
    axis_font = {'fontname' : 'Arial', 'size':'16', 'weight':'bold'}
    colors    = ('b', 'g', 'r', 'c', 'm', 'y', 'k', 'w')

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    ax.tick_params(axis='both', which='minor', labelsize=14)
    ax.tick_params(axis='both', which='major', labelsize=14)

    ax.axvline(x=timeout, color='r', zorder=3, linestyle='--')
    ax.axhline(y=timeout, color='r', zorder=3, linestyle='--')
    c = np.arange(0, timeout, 0.1)
    ax.plot(c, c, c='r', zorder=3, linestyle='--')

    scat = ax.scatter(x, y, c=z, cmap=plt.cm.brg, marker='x', linewidth=1, s=150, zorder=2)
    #fig.colorbar(scat, orientation='vertical', label=z_label) # unbounded density bar

#    fit = np.polyfit(np.log10(x), np.log10(y), 1)
#    fit_fn = poly1d(fit)
#    ax.plot(c, 10**fit_fn(np.log10(c)), '--g')

    ax.grid(True)
    xlim([0.005, timeout*1.2])
    ylim([0.005, timeout*1.2])

#    ax.set_title(title, **axis_font)
    ax.set_ylabel(y_label, **axis_font)
    ax.set_yscale('log')#, nonposy='clip')
    ax.set_xlabel(x_label, **axis_font)
    ax.set_xscale('log')#, nonposy='clip')
    ax.set_aspect('equal')  # does not truly work! 

    # dump to file
    plt.tight_layout()
    plt.savefig(file_name, bbox_inches='tight')
#   plt.show()
    plt.close(fig)

def get_cactus_data(configs, results, timeout):
    data = []

    for config in configs:
        v = []
        solved = 0
        for bench in results.keys():
            res = results[bench][config]
            if res[1] == 0 and res[0] < timeout:
                v.append(res[0])
                solved += 1
#            else:
#                v.append(600.0)
#                solved += 0
        v.sort()
        v = np.cumsum(v)

        config_data = {}
        config_data["config"] = config
        config_data["cum_times"] = v
        config_data["max"] = max(v)
        config_data["solved"] = solved
        data.append(config_data)

    v = []
    solved = 0

    for bench in results.keys():
        best_time = None
        best_config = None
        for config in configs:
            res = results[bench][config]
            if res[1] == 0 and float(res[0]) < float(timeout):
                if best_time is None:
                    best_time = res[0]
                    best_config = config
                elif float(res[0]) < float(best_time):
                    best_time = res[0]
                    best_config = config

        if best_time is not None:
            v.append(best_time)
            solved += 1
#        else:
#            v.append(600.0)
#            solved += 0
    v.sort()
    v = np.cumsum(v)

    config_data = {}
    config_data["config"] = "virtual best"
    config_data["cum_times"] = v
    config_data["max"] = max(v)
    config_data["solved"] = solved
    data.append(config_data)

    print(config_data)

    data.sort(key = lambda x : x["max"])
    data.sort(key = lambda x : x["solved"], reverse=True)
    return data

def save_cactus_plots(plots_dir, configs, results, timeout):
        title   = "Tools Performance"
        x_label = "Solved Instances" # "No. solved"
        y_label = "Cumulative time (s.)" #"Cum. time (s.)"
        x_label = "No. solved"
        y_label = "Cum. time (s.)"
        data      = get_cactus_data(configs, results, timeout)

        configs += "virtual best"
        
        file_name = "%s/cactus_plot.png" % plots_dir
        plot_cactus_plot(title, x_label, y_label, data, timeout, file_name)

        file_name = "%s/cactus_plot.pdf" % plots_dir
        plot_cactus_plot(title, x_label, y_label, data, timeout, file_name)

def plot_cactus_plot(title, x_label, y_label, data, timeout, file_name):
    axis_font = {'fontname' : 'Arial', 'size' : '14', 'weight': 'bold'}
    std_colors    = ('b', 'g', 'r', 'c', 'm', 'y', 'k', 'w')
    num_keys  = len(data)
    cmap      = plt.get_cmap('Dark2_r')
    cmap      = plt.cm.nipy_spectral
    colors    = [cmap(i) for i in np.linspace(0, 1, num_keys)]
    styles    = ('--', '-.', '-', ':')

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    idx = 0
    max_y = 0
    for config_data in data:
        y = config_data["cum_times"]
        x = np.arange(1, len(y) + 1)
        if "ofpbs" in config_data["config"]:
            pass
        if "omt" in config_data["config"]:
            pass
        ax.plot(x, y, styles[idx % len(styles)], label=get_pretty_label(config_data["config"]), marker=None, color=colors[idx], linewidth=1)
        idx += 1
        if config_data["max"] > max_y:
            max_y = config_data["max"]

    ax.grid(True)
#    ax.set_title(title, **axis_font)
    ax.set_ylabel(y_label, **axis_font)
    ax.set_xlabel(x_label, **axis_font)
    ax.set_yscale('log', nonposy='clip')
    plt.ylim([0, max_y * 1.5])

    lgd = plt.legend(loc=5,ncol=2, bbox_to_anchor=(1.000, 0.19), prop={'size' : 6})
#    lgd = plt.legend(loc=5,ncol=2, bbox_to_anchor=(1.0, 0.2))
#    lgd = plt.legend(loc=5,ncol=2, bbox_to_anchor=(0.99, -0.4))

    plt.savefig(file_name, bbox_extra_artists=(lgd,), bbox_inches='tight')#, bbox_inches=0)
    plt.close(fig)


def get_pretty_label(name):
    res = "OptiM.("
    if "ofpbs" in name:
        res += "eager+ofpbs"
        if "ttt" in name:
            res += "+bp+pi+so"
        elif "ttf" in name:
            res += "+bp+pi"
        elif "tft" in name:
            res += "+pi+so"
        elif "tff" in name:
            res += "+pi"
        elif "ftt" in name:
            res += "+bp+so"
        elif "ftf" in name:
            res += "+bp"
        elif "fft" in name:
            res += "+so"
        elif "fff" in name:
            res += ""
        res += ")"
    elif "omt" in name:
        if "_e_" in name:
            res += "eager"
        elif "_l_" in name:
            res += "lazy"
        res += "+omt"
        if "bin" in name:
            res += "+bin"
        else:
            res += "+lin"
        if "ttf" in name:
            res += "+bp+pi"
        elif "tff" in name:
            res += "+pi"
        elif "ftf" in name:
            res += "+bp"
        elif "fff" in name:
            res += ""
        res += ")"
    elif "e_r_obvbs" in name:
        res += "obvbs) [reduction]"
    elif "smt" in name:
        res += "smt"
        res += ")"
    else:
        res = name
    return res

def mkdir_p(path):
    """make a directory and any missing ancestor if needed"""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



###
###
###

if (__name__ == "__main__"):
    main()

